import os
import ast
import types

from monty.termcolor import cprint
from abipy.flowtk import Flow


def as_orderedset(token, options):
    """
    Helper function used to parse --mpi-list argument.
    Return OrderedSet
    """

    if token.startswith("range"):
        # start(1,4,2)
        token = token[5:]
        #print("token", token)
        t = ast.literal_eval(token)
        assert len(t) == 3
        l = list(range(t[0], t[1], t[2]))

    elif token.endswith("x"):
        # 16x --> multiple of 16
        fact = int(token[:-1])
        l, i = [], 0
        while True:
            i += 1
            val = fact * i
            if val > options.max_ncpus: break
            l.append(val)

    elif token.startswith("pow"):
        # pow4 --> powers of 4
        base = int(token[3:])
        l, i = [], 0
        while True:
            i += 1
            val = base ** i
            if val > options.max_ncpus: break
            l.append(val)
    else:
        # lists
        l = ast.literal_eval(token)

    #print("l", l)
    l = [n for n in l if options.min_ncpus <= n <= options.max_ncpus]
    from orderedset import OrderedSet
    return OrderedSet(l)


class BenchmarkFlow(Flow):
    """
    Flow used to benchmark Abinit.
    Add useful methods to post-process the results.
    """

    def exclude_from_benchmark(self, node):
        """
        Exclude a task or the tasks in a Work from the benchmark analysis.
        """
        if not hasattr(self, "_exclude_nodeids"): self._exclude_nodeids = set()

        if node.is_work:
            for task in node:
                self._exclude_nodeids.add(task.node_id)
        else:
            assert node.is_task
            self._exclude_nodeids.add(node.node_id)

    @property
    def exclude_nodeids(self):
        if not hasattr(self, "_exclude_nodeids"): self._exclude_nodeids = set()
        return self._exclude_nodeids

    def get_parser(self):
        """
        Parse the timing sections in the output files.
        Return AbinitTimerParser parser object for further analysis.
        """
        nids = []
        for task in self.iflat_tasks():
            if task.node_id in self.exclude_nodeids: continue
            if task.status != task.S_OK: continue
            #print("analysing task:", task)
            nids.append(task.node_id)

        parser = self.parse_timing(nids=nids)

        if parser is None:
            print("flow.parse_timing returned None!")
        else:
            if len(parser) != len(nids):
                print("Not all timing sections have been parsed!")

        return parser

    def build_and_pickle_dump(self, **kwargs):
        cnt = 0
        for task in self.iflat_tasks():
            if task.node_id in self.exclude_nodeids: continue
            cnt += 1
            #print("%s: mpi_procs %d, omp_threads %d" %
            #  (task, task.manager.qadapter.mpi_procs, task.manager.qadapter.omp_threads))

        print("Total number of benchmarks: %d" % cnt)
        return super().build_and_pickle_dump(**kwargs)

    #def make_tarball(self):
    #    self.make_tarfile(self, name=None, max_filesize=None, exclude_exts=None, exclude_dirs=None, verbose=0, **kwargs):


def bench_main(main):
    """
    This decorator is used to decorate main functions producing `AbinitFlows`.
    It adds the initialization of the logger and an argument parser that allows one to select
    the loglevel, the workdir of the flow as well as the YAML file with the parameters of the `TaskManager`.
    The main function shall have the signature:

        main(options)

    where options in the container with the command line options generated by `ArgumentParser`.

    Args:
        main:
            main function.
    """
    from functools import wraps

    @wraps(main)
    def wrapper(*args, **kwargs):  # pragma: no cover
        parser = build_bench_main_parser()
        options = parser.parse_args()

        # loglevel is bound to the string value obtained from the command line argument.
        # Convert to upper case to allow the user to specify --loglevel=DEBUG or --loglevel=debug
        import logging
        numeric_level = getattr(logging, options.loglevel.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: %s' % options.loglevel)
        logging.basicConfig(level=numeric_level)

        bench_monkey_patch_options(options)

        # Istantiate the manager.
        from abipy.abilab import TaskManager
        options.manager = TaskManager.as_manager(options.manager)

        #if options.tempdir:
        #    import tempfile
        #    options.workdir = tempfile.mkdtemp()
        #    print("Working in temporary directory", options.workdir)

        flow = main(options)
        if flow is None: return 0

        if options.abivalidate:
            print("Validating input files of the flow...")
            isok, errors = flow.abivalidate_inputs()
            if not isok:
                for e in errors:
                    if e.retcode == 0: continue
                    lines = e.log_file.readlines()
                    i = len(lines) - 50 if len(lines) >= 50 else 0
                    print("Last 50 line from logfile:")
                    print("".join(lines[i:]))
                raise RuntimeError("flow.abivalidate_input failed. See messages above.")
            else:
                print("Validation succeeded")
                return 0

        #if options.remove and os.path.isdir(options.workdir):
        #    print("Removing old directory:", options.workdir)
        #    import shutil
        #    shutil.rmtree(options.workdir)

        if options.scheduler:
            return flow.make_scheduler().start()
        else:
            return flow.build_and_pickle_dump()

    return wrapper


def build_bench_main_parser():
    """Build the parser used in the abipy/benchmarks scripts"""
    import argparse
    parser = argparse.ArgumentParser()

    parser.add_argument('--loglevel', default="ERROR", type=str,
                        help="set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")

    parser.add_argument('-v', '--verbose', default=0, action='count', # -vv --> verbose=2
                              help='verbose, can be supplied multiple times to increase verbosity')

    parser.add_argument("-w", '--workdir', default="", type=str, help="Working directory of the flow.")

    parser.add_argument("-m", '--manager', default=None,
                        help="YAML file with the parameters of the task manager. "
                             "Default None i.e. the manager is read from standard locations: "
                             "working directory first then ~/.abinit/abipy/manager.yml.")

    parser.add_argument("--mpi-list", default=None, type=str, help="List of MPI processors to be tested. Syntax:\n"
                        "--mpi-list=[1,6,12] to define a list, range(1,4,2) for a python range.\n"
                        "--mpi-list=16x for multiple of 16 up to max--ncpus, --mpi-list=pow2 for powers of 2")
    parser.add_argument("--omp-list", default=None, type=str,
                        help="List of OMP threads to be tested. Default is [1]. Same syntax as mpi-list.")

    parser.add_argument("--min-ncpus", default=-1, type=int, help="Minimum number of CPUs to be tested.")
    parser.add_argument("--max-ncpus", default=248, type=int, help="Maximum number of CPUs to be tested. Default: 248.")
    parser.add_argument("--min-eff", default=None, type=float, help="Minimum parallel efficiency accepted. Default None.")

    parser.add_argument('--paw', default=False, action="store_true", help="Run PAW calculation if available")
    parser.add_argument("-a", '--abivalidate', default=False, action="store_true",
                        help="Call Abinit to validate input files and return")

    parser.add_argument("-i", '--info', default=False, action="store_true", help="Show benchmark info and exit")
    parser.add_argument("-r", "--remove", default=False, action="store_true", help="Remove old flow workdir")

    parser.add_argument("--scheduler", "-s", default=False, action="store_true", help="Run with the scheduler")
    return parser


def bench_monkey_patch_options(options):
    """Monkey patch options to add useful method"""
    # parse arguments
    if options.mpi_list is not None:
        options.mpi_list = as_orderedset(options.mpi_list, options)

    if options.omp_list is None:
        options.omp_list = [1]
    else:
        options.omp_list = as_orderedset(options.omp_list, options)

    # Monkey patch options to add useful method
    def _monkey_patch(opts):
        # options.accept_mpi_omp(mpi_proc, omp_threads)
        def accept_mpi_omp(opts, mpi_procs, omp_threads):
            """Return True if we can run a benchmark with mpi_procs and omp_threads"""
            tot_ncpus = mpi_procs * omp_threads

            if tot_ncpus < opts.min_ncpus:
                if options.verbose:
                    cprint("Skipping %d because of min_ncpus" % tot_ncpus, color="magenta")
                return False

            if opts.max_ncpus is not None and tot_ncpus > opts.max_ncpus:
                if options.verbose:
                    cprint("Skipping %d because of max_ncpus" % tot_ncpus, color="magenta")
                return False

            return True

        opts.accept_mpi_omp = types.MethodType(accept_mpi_omp, opts)

        def accept_conf(opts, conf, omp_threads):
            """Return True if we can run a benchmark with mpi_procs and omp_threads"""
            tot_ncpus = conf.mpi_procs * omp_threads

            if tot_ncpus < opts.min_ncpus:
                if options.verbose:
                    cprint("Skipping %d because of min_ncpus" % tot_ncpus, color="magenta")
                return False

            if opts.max_ncpus is not None and tot_ncpus > opts.max_ncpus:
                if options.verbose:
                    cprint("Skipping %d because of max_ncpus" % tot_ncpus, color="magenta")
                return False

            if opts.min_eff is not None and conf.efficiency < opts.min_eff:
                if options.verbose:
                    cprint("Skipping %d because of parallel efficiency" % tot_ncpus, color="magenta")
                return False

            if options.verbose:
                cprint("Accepting omp_threads:%s with conf\n%s" % (omp_threads, conf), color="green")

            return True

        opts.accept_conf = types.MethodType(accept_conf, opts)

        # options.get_workdir(__file__)
        def get_workdir(opts, _file_):
            """
            Return the workdir of the benchmark.
            A default value if constructed from the name of the scrip if no cmd line arg.
            """
            if options.workdir: return options.workdir
            return "bench_" + os.path.basename(_file_).replace(".py", "")

        opts.get_workdir = types.MethodType(get_workdir, opts)

    _monkey_patch(options)



#!/usr/bin/env python
r"""
Potentials
==========

This example shows how to plot the potentials stored in netcdf files.
Use the input variables `prtpot`, `prtvha`, `prtvhxc`, `prtvxc` with
`iomode 3` to produce these files at the end of the SCF-GS run.
"""
from abipy.abilab import abiopen
import abipy.data as abidata

# VKS = Hartree + XC potential + sum of local part of pseudos.
with abiopen(abidata.ref_file("ni_666k_POT.nc")) as ncfile:
    vks = ncfile.vks

#vks.plot_line(point1=[0, 0, 0], point2=[0, 4, 0], num=400, title="$V_{ks}(r)$")

# Hartree potential.
with abiopen(abidata.ref_file("ni_666k_VHA.nc")) as ncfile:
    vh = ncfile.vh

vh.plot_line(point1=[0, 0, 0], point2=[0, 4, 0], num=400, title="$V_{hartree}(r)$")

# XC potential.
with abiopen(abidata.ref_file("ni_666k_VXC.nc")) as ncfile:
    vxc = ncfile.vxc

vxc.plot_line(point1=[0, 0, 0], point2=[0, 4, 0], num=400, title="$V_{xc}(r)$")

# Hartree + XC potential.
with abiopen(abidata.ref_file("ni_666k_VHXC.nc")) as ncfile:
    vhxc = ncfile.vhxc

vhxc.plot_line(point1=[0, 0, 0], point2=[0, 4, 0], num=400, title="$V_{hxc}(r)$")

vloc = vks - vhxc
vloc.plot_line(point1=[0, 0, 0], point2=[0, 4, 0], num=400, title="$V_{loc}(r)$")

foo = vhxc - vh - vxc
#foo.plot_line(point1=[0, 0, 0], point2=[0, 4, 0], num=400, title="$V_{hxc - h - xc}(r)$")

# To plot the wavefunction along the lines connect the firt atom in the structure
# and all the neighbors within a sphere of radius 3 Angstrom:
#vxc.plot_line_neighbors(site_index=0, radius=3)



#!/usr/bin/env python
"""Benchmark for phonon calculation with DFPT."""
import sys
import numpy as np
import abipy.abilab as abilab
import abipy.flowtk as flowtk
import abipy.data as abidata

from itertools import product
from abipy.benchmarks import bench_main, BenchmarkFlow


def make_inputs(paw=False):
    pseudos = abidata.pseudos("56ba.psp_mod", "22ti.psp_mod", "8o.psp_mod")
    #pseudos = abidata.pseudos("56ba.psp_mod", "22ti.psp_mod", "8o.pspnc")
    if paw: raise NotImplementedError("PAW")

    # SLAB ending TiO2 double layer. paralelectric configuration
    # N=9
    # Supercell and atoms

    xcart = np.fromstring("""
0.0000000000E+00  0.0000000000E+00 -4.2633349730E+00
3.7794522658E+00  3.7794522658E+00 -3.2803418097E+00
0.0000000000E+00  3.7794522658E+00 -3.6627278067E+00
3.7794522658E+00  0.0000000000E+00  6.5250113947E-01
0.0000000000E+00  3.7794522658E+00 -1.0555036964E-01
3.7794522658E+00  3.7794522658E+00  3.2682166278E-01
0.0000000000E+00  0.0000000000E+00  3.9815918094E+00
3.7794522658E+00  3.7794522658E+00  4.0167907030E+00
3.7794522658E+00  0.0000000000E+00  7.7541444349E+00
0.0000000000E+00  3.7794522658E+00  7.6664087705E+00
3.7794522658E+00  3.7794522658E+00  7.7182324796E+00
0.0000000000E+00  0.0000000000E+00  1.1412913350E+01
3.7794522658E+00  3.7794522658E+00  1.1416615533E+01
3.7794522658E+00  0.0000000000E+00  1.5117809063E+01
0.0000000000E+00  3.7794522658E+00  1.5117809063E+01
3.7794522658E+00  3.7794522658E+00  1.5117809063E+01
0.0000000000E+00  0.0000000000E+00  1.8822704777E+01
3.7794522658E+00  3.7794522658E+00  1.8819002593E+01
3.7794522658E+00  0.0000000000E+00  2.2481473692E+01
0.0000000000E+00  3.7794522658E+00  2.2569209355E+01
3.7794522658E+00  3.7794522658E+00  2.2517385647E+01
0.0000000000E+00  0.0000000000E+00  2.6254026317E+01
3.7794522658E+00  3.7794522658E+00  2.6218827423E+01
3.7794522658E+00  0.0000000000E+00  2.9583116986E+01
0.0000000000E+00  3.7794522658E+00  3.0341168496E+01
3.7794522658E+00  3.7794522658E+00  2.9908796464E+01
0.0000000000E+00  0.0000000000E+00  3.4498953099E+01
3.7794522658E+00  3.7794522658E+00  3.3515959935E+01
0.0000000000E+00  3.7794522658E+00  3.3898345933E+01
""", sep=" ").reshape((-1,3))

    # Crystal structure.
    structure = abilab.Structure.from_abivars(
        #acell="4.0 4.0 28.0 Angstrom",
        acell=abilab.ArrayWithUnit([4.0, 4.0, 28], "ang").to("bohr"),
        rprim=np.eye(3),
        typat=[int(i) for i in "3 3 2 3 3 2 1 3 3 3 2 1 3 3 3 2 1 3 3 3 2 1 3 3 3 2 3 3 2".split()],
        znucl=[56, 22, 8],
        xcart=xcart,
    )

    multi = abilab.MultiDataset(structure, pseudos, ndtset=2)

    # Global variables used both for the GS and the DFPT run.
    multi.set_vars(
        #electronic structure
        nband=120,
        ecut=15.0,
        pawecutdg=30.0 if paw else None,
        nstep=80,
        ngkpt=[4,4,1],
        shiftk=[0,0,0],
        paral_kgb=1,
        timopt=-1,
        prtden=0,
    )

    gs_inp, ph_inp = multi.split_datasets()

    gs_inp.set_vars(tolvrs=1e-6, kptopt=1)

    ph_inp.set_vars(
        kptopt=3,
        rfphon=1,
        irdwfk=1,
        rfatpol=[1, 1],
        rfdir=[0, 0, 1],
        nqpt=1,
        qpt=[0.0, 0.25, 0.0],
        prtwf=0,
        #tolwfr=1.0e-22,
        toldfe=1.0e-9,
        tolrde=0.0, # This is a development input variable, used in the present case to avoid load unbalance
                    # when studying the scaling with respect to the number of cores. Do not define it in
                    # your production runs
    )

    return gs_inp, ph_inp


def build_flow(options):
    gs_inp, ph_inp = make_inputs()

    flow = BenchmarkFlow(workdir=options.get_workdir(__file__), remove=options.remove)
    gs_work = flowtk.Work()
    gs_work.register_scf_task(gs_inp)
    flow.register_work(gs_work)
    flow.exclude_from_benchmark(gs_work)

    # Get the list of possible parallel configurations from abinit autoparal.
    max_ncpus, min_eff = options.max_ncpus, options.min_eff
    print("Getting all autoparal confs up to max_ncpus:", max_ncpus," with efficiency >=", min_eff)

    pconfs = ph_inp.abiget_autoparal_pconfs(max_ncpus, autoparal=1)
    if options.verbose: print(pconfs)

    work = flowtk.Work()
    for conf, omp_threads in product(pconfs, options.omp_list):
        if not options.accept_conf(conf, omp_threads): continue

        manager = options.manager.new_with_fixed_mpi_omp(conf.mpi_procs, omp_threads)
        inp = ph_inp.new_with_vars(conf.vars)
        work.register_phonon_task(inp, manager=manager, deps={gs_work[0]: "WFK"})

    print("Found %d configurations" % len(work))
    flow.register_work(work)

    return flow.allocate()


@bench_main
def main(options):
    if options.info:
        # print doc string and exit.
        print(__doc__)
        return
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())
